% Copyright (C) Paweł Płazieński 2009
%
% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License as published by the
% Free Software Foundation; either version 2, or (at your option) any
% later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

propertize(List,Property) :-
	sort(List,P),
	Property = property(P).

property_to_list(Property,List) :-
	Property = property(List).

add_property(Property,DependOnList,DependForList) :-
	model(Model),
	history(History),
	add_property(Model,History,Property,DependOnList,DependForList,NewModel,NewHistory),
	model_set(NewModel),
	history_set(NewHistory).

add_dependancy(PropertyFrom,PropertyTo) :-
	model(Model),
	add_dependancy(Model,PropertyFrom,PropertyTo,NewModel),
	model_set(NewModel).

finalize(OldProperty,NewProperty) :-
	model(Model),
	history(History),
	finalize(Model,History,OldProperty,NewProperty,NewModel,NewHistory),
	model_set(NewModel),
	history_set(NewHistory).

split(OldProperty,NewProperty) :-
	model(Model),
	history(History),
	split(Model,History,OldProperty,NewProperty,NewModel,NewHistory),
	model_set(NewModel),
	history_set(NewHistory).

get_properties(List) :-
	model(Model),
	vertices(Model,List).

get_simple_properties(List) :-
	get_properties(All),
	setof(X, (member(X,All), simple_property(X)), List), !.

get_simple_properties([]).

simple_property(X) :-
	X = property([_]).

properties_to_strings([],[]).

properties_to_strings(Properties,Strings) :-
	Properties = [PropertiesHead|PropertiesTail],
	Strings = [StringsHead|StringsTail],
	properties_to_strings(PropertiesTail,StringsTail),
	property_to_string(PropertiesHead,StringsHead).

property_to_string(Property,String) :-
	Property = property(List),
	list_string(List,String,',').

% private

add_property(Model,History,Property,DependBackList,DependForwList,NewModel,NewHistory) :-
	add_dependancy(Model,DependBackList,Property,DependBackModel),
	add_dependancy(DependBackModel,Property,DependForwList,DependForwModel),
	add_vertices(History,[Property],NewHistory),
	add_vertices(DependForwModel,[Property],NewModel).

add_dependancy(Model,Predecessor,Successor,NewModel) :-
	Predecessor = property(_),
	Successor = property(_),
	add_edges(Model,[Predecessor-Successor],NewModel).

add_dependancy(Model,_,[],Model).

add_dependancy(Model,Predecessor,[SuccessorHead|SuccessorTail],NewModel) :-
	add_dependancy(Model,Predecessor,SuccessorHead,HeadModel),
	add_dependancy(HeadModel,Predecessor,SuccessorTail,NewModel).

add_dependancy(Model,[],_,Model).

add_dependancy(Model,[PredecessorHead|PredecessorTail],Successor,NewModel) :-
	add_dependancy(Model,PredecessorHead,Successor,HeadModel),
	add_dependancy(HeadModel,PredecessorTail,Successor,NewModel).

replace(OldSet,OldElement,NewElementList,NewSet) :-
	subtract(OldSet,[OldElement],Short),
	(  OldSet = Short,
		NewSet = OldSet
	;  not(OldSet = Short),
		union(NewElementList,Short,NewSet)
	).

edge_replace([],_,_,[]).

edge_replace(OldEdges,OldProperty,NewPropertyList,NewEdges) :-
	OldEdges = [OldEdgesHead|OldEdgesTail],
	edge_replace(OldEdgesTail,OldProperty,NewPropertyList,NewEdgesTail),
	OldEdgesHead = OldEdgeSource-OldEdgeTarget,
	(  (OldEdgeSource = OldProperty, OldEdgeTarget = OldProperty),
		setof(N-N, member(N,NewPropertyList), NewEdgesUnion)
	;  (not(OldEdgeSource = OldProperty), OldEdgeTarget = OldProperty),
		setof(OldEdgeSource-N, member(N,NewPropertyList), NewEdgesUnion)
	;  (OldEdgeSource = OldProperty, not(OldEdgeTarget = OldProperty)),
		setof(N-OldEdgeTarget, member(N,NewPropertyList), NewEdgesUnion)
	;  (not(OldEdgeSource = OldProperty), not(OldEdgeTarget = OldProperty)),
		NewEdgesUnion = [OldEdgesHead]
	),
	append(NewEdgesTail,NewEdgesUnion,NewEdges).
model_replace(Model,OldProperty,NewPropertyList,NewModel) :-
	vertices(Model,OldVertices),
	replace(OldVertices,OldProperty,NewPropertyList,NewVertices),
	edges(Model,OldEdges),
	edge_replace(OldEdges,OldProperty,NewPropertyList,NewEdges),
	vertices_edges_to_ugraph(NewVertices,NewEdges,NewModel).

finalize(Model,History,OldProperty,NewProperty,NewModel,NewHistory) :-
	add_edges(History,[OldProperty-NewProperty],NewHistory),
	model_replace(Model,OldProperty,[NewProperty],NewModel).

split(Model,History,OldProperty,NewPropertyList,NewModel,NewHistory) :-
	ugraph_union(History,[OldProperty-NewPropertyList],NewHistory),
	% TODO something more obvious to reader
	model_replace(Model,OldProperty,NewPropertyList,NewModel).

:- dynamic(history/1).
:- dynamic(model/1).

history_set(NewHistory) :-
	retractall(history(_)),
	assert(history(NewHistory)).

model_set(NewModel) :-
	retractall(model(_)),
	assert(model(NewModel)).

clear_data :-
	retractall(history(_)),
	retractall(model(_)),
	assert(history([])),
	assert(model([])).

:- assert(history([])).
:- assert(model([])).


